#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <io.h>
#include <sys/stat.h>

#include "resource.h"
#include "shared.h"

#define	DBGLOG			1		/// set to 1 to log all of our I/O to BuddyLog, set 0 to ...uh... NOT log it
#define	TRACELEN		128		/// how many opcode addresses to buffer

/*

NOTE: A Buddy can be written to follow various constraints with regard to the Emulator:

1) It can require that Emulator be run first. If, when Buddy starts up, pMailBox doesn't exist, or pMailBox->hWnd[0]==0, then Buddy terminates.
2) It can require that Buddy be run first. If, when starting up, pMailBox exists AND pMailBox->hWnd[0]!=0, then Buddy terminates.
3) It can auto-terminate when Emualator terminates. When BMSG_GONE is received from Emulator, Buddy terminates itself.
4) At some point, Emulator may be modified to all its configuration to auto-load/start a list of Buddies.
5) A Buddy window may be visible or hidden. Just leave (or omit) the ShowWindow() call in WinMain().

*/

LRESULT CALLBACK MainProc(HWND, UINT, WPARAM, LPARAM);

char		szAppName[] = "Series 80 Emulator Buddy CPM";
HANDLE		hInstance;
HWND		hWndMain;
char		ScratBuf[8192];

DWORD		SlotNum=3;						/// S.C. for Interrupt purposes (probably not used for CP/M)
DWORD		CCRADR=0177506, IBADR=0177507;	/// our CPM Card I/O addresses
#define		PSRADR		CCRADR
#define		OBADR		IBADR
BYTE		CCR, PSR, IB, OB;

/*

Z80 pins of interest:

RESET - pin 26
INT - interrupt line pin 16
NMI - non-maskable-interrupt line pin 17 (tied high)
WAIT - pin 24 (tied high)
HALT - pin 18 (open)

Z80 registers:

IX
IY
SP	points TO the top-most byte on the stack, NOT one higher! AND, the stack grows downward! push/pop is always 2-byte op
PC

7 A
A'

0 B
1 C

B'
C'

2 D
3 E

D'
E'

4 H
5 L

6 memory

H'
L'
I interrupt-page address register (high-order byte of the interrupt address, lower part is provided by the device generating the interrupt
Flags

instructions can be from 1 to 4 bytes.
*/

#define		IOPS_WANT		0	/// need to send BMSG_WANT and move on to IOPS_HOOK
#define		IOPS_HOOK		1	/// need to ask for available I/O slots, pick one, hook it, and move on to IOPS_IDLE
#define		IOPS_IDLE		2	/// normal state to hang out in
#define		IOPS_RESET1		3	/// RESET has happened, respond with INT and send first 3 (self-test OK)
#define		IOPS_RESET2		4	/// respond with second 3 (self-test OK)
#define		IOPS_INT		5	/// CPU80 has placed us in an INTERRUPTED state to process commands

int			IOPstate=IOPS_WANT;

#if DBGLOG
///*************************************************************************************************************************************************************
/// DBGLOG stuff enables TRACING of CPU80 instructions in bank-swapped ROMs (to know where I/O accesses are coming from) and LOGGING of I/O accesses to BuddyLog
///*************************************************************************************************************************************************************
HANDLE		hLogFile=NULL;
WORD		Trace[TRACELEN];	/// a rolling wrap-around buffer to keep track of bank-switched ROM addresses that access our I/O ports
int			NextTrace=0;		/// the next location in Trace[] to store the address of an I/O port access
BYTE		TraceROM;			/// the ROM# of the ROM which LAST accessed our I/O ports
///******************************************************************************************************
void OpenLogFile() {

	if( hLogFile==NULL ) {
		hLogFile = CreateFile("BuddyLog", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if( hLogFile==INVALID_HANDLE_VALUE ) hLogFile = NULL;
		else SetFilePointer(hLogFile, 0, NULL, FILE_END);	// seek to end of file
	}
}

///********************************************
void Log() {

	int		t, i;
	char	*p;
	DWORD	len;

	if( hLogFile ) {
		sprintf(ScratBuf+strlen(ScratBuf), " ROM:%03o ", TraceROM);

		t = 0;
		i = NextTrace-1;
		p = ScratBuf+strlen(ScratBuf);

		do {
			if( i<0 ) i = TRACELEN-1;
			sprintf(p, " %07o", Trace[i--]);
			p += 8;
		} while( ++t < TRACELEN );

		strcat(ScratBuf, "\r\n");
		WriteFile(hLogFile, ScratBuf, strlen(ScratBuf), &len, NULL);
	}
}

///*******************************************
void CloseLogFile() {

	if( hLogFile!=NULL ) CloseHandle(hLogFile);
	hLogFile = NULL;
}
///*************************************************************************************************************************************************************
/// DBGLOG stuff end
///*************************************************************************************************************************************************************
#endif

///******************************************
void GenInt() {

	/// assumes, if we're interrupting CapCPU, that we haven't already got a message pending
	/// and even if we do, we overwrite it, otherwise deadlock looms over our shoulder...
	pMailBox->ArgB2E[MailBoxNum][0] = SlotNum;
	pMailBox->MsgB2E[MailBoxNum] = BMSG_INT;
}

///*******************************************
/// We need to close/unhook everything and reset state
void CloseShop() {

#if DBGLOG
	CloseLogFile();
#endif

#if SERIAL_FILE
	ClosePrintFile();
#endif

	/// release our 'taken' I/O addresses
	if( pMailBox->hWnd[0] && IOPstate>=IOPS_HOOK ) {	/// if Emulator is still there and we're hooked into it
		memset((void*)&pMailBox->ArgB2E[MailBoxNum][0], 0, sizeof(pMailBox->ArgB2E[MailBoxNum]));	/// clear all the possible I/O requests
		*(((WORD*)pMailBox->ArgB2E[MailBoxNum])+0) = CCRADR;	/// add the list of I/O addresses we 'took'
		*(((WORD*)pMailBox->ArgB2E[MailBoxNum])+1) = IBADR;
		pMailBox->MsgB2E[MailBoxNum] = BMSG_IOREL;				/// ask Emulator to release our hooks into those I/O address
		while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum] ) ;	/// as long as Emulator doesn't terminate, wait until message is processed
	}
	IOPstate = IOPS_WANT;			/// Emulator called it quits, so reset everything, in case we're sticking around
}

///******************************************************************************************************
/// check OBF, deal with it
void HandleOBF() {

	if( PSR & 0x80 ) {	/// if OBF
		// from the CPM rev C BPGM, it appears that the 87 side only writes 200 or 000 to CCR (to reset the Z80)
		// and only checks the MS (OBF) and LS (IBF) bits of PSR. So, that simplifies this Buddy's interface code significantly
		// compared to the standard I/O Interface cards.

		PSR &= ~0x80;		/// indicate we've read the OB, so it's now empty
	}
}

///******************************************************************************************************
int WINAPI WinMain (HINSTANCE hThisInstance, HINSTANCE hPrevInstance, LPSTR lpszArgument, int nCmdShow) {

    MSG			msg;
    WNDCLASS	wc;		// Data structure for the windowclass
    int			n, i;
    DWORD		adr;
    BYTE		b;

    hInstance = hThisInstance;

    // build the windows class structure
    wc.lpfnWndProc = MainProc;				// This function is called by windows
    wc.cbClsExtra = 0;						// No extra bytes after the window class
    wc.cbWndExtra = 0;						// structure or the window instance
    wc.hInstance = hInstance;
    wc.hIcon = LoadIcon (hInstance, MAKEINTRESOURCE(IDI_ICON1));
    wc.hCursor = LoadCursor (NULL, IDC_ARROW);
    wc.lpszMenuName = NULL;					// No menu
    wc.lpszClassName = szAppName;
    wc.style = CS_OWNDC | CS_VREDRAW | CS_HREDRAW;
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);

    // Register the window class, and if it fails quit the program
    if( !RegisterClass(&wc) ) return 0;

	hWndMain = CreateWindow(szAppName, szAppName, WS_OVERLAPPEDWINDOW,
		475, 10, 544, 375, NULL, NULL, hInstance, NULL);

    /// setup the share link
    StartShare();
    if( pMailBox==NULL ) {
		MessageBox(NULL, "Unable to open MAILBOX", szAppName, MB_OK);
		return 0;
    }
    /// find and take an open mailbox
    for(n=1; n<8; n++) {
		if( pMailBox->hWnd[n]==0 ) {
			MailBoxNum = n;
			pMailBox->hWnd[MailBoxNum] = hWndMain;		// Emulator is ALWAYS MAILBOX[0]
			break;
		}
    }
    if( n>=8 ) {	// too many Buddies running, bail out (unlikely, but best to check...)
		EndShare(FALSE);
		return 0;
    }

	/// NOTE: This shows the window. If you don't want a window to show on the screen for your Buddy, comment out the following line.
    ShowWindow(hWndMain, nCmdShow);

    /// =========================================
    /// Windows message pump AND MailBox monitor
    /// =========================================
	while( TRUE ) {
		/// Windows messages take priority, to keep the window responsive.

		/// When there's no Windows message, then we check for Emulator messages.

		///======================================
		/// check if Windows message
		///======================================
		/// we first check if we have a Windows message, before calling GetMessage(),
		/// as otherwise, if we call GetMessage() when there's NO message,
		/// it won't come back until there IS a message, and that will
		/// block our communications between Emulator and Buddy.
		while( !PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) ) {
			///========================================
			/// check for Emulator message
			///========================================

			/// NOTE: In this sample CP/M emulation, we're sending a BWANT_TRACE so that
			/// we can trace the Series 80 I/O addresses that were executing leading up to each
			/// I/O access to our CP/M I/O addresses. That makes debugging and figuring out, what the
			/// system BPGM sends to the Serial Card and what it expects back, much easier.
			/// In a real, 'production' buddy, the BWANT_TRACE would not be done, as we have no other
			/// need for trapping all those instructions, which process slows down the emulation noticeably.

			/// NOTE: I/O interface cards are capable of HALTing the Series 80 CPU by setting the MSbit
			/// of the CCR/PSR register. That bit gets mapped by the Translator IC to pull on the HALTEN
			/// line that freezes the Series 80 CPU in its tracks. I don't know whether any of the extant
			/// I/O interface cards make use of that feature, but should YOU want to emulate that in a
			/// Buddy, instead of setting the MSbit in CCR/PSR, you should send a BMSG_HALT with
			/// pMailBox->ArgB2E[MailBoxNum][0]=1 and Emulator will 'hang' in a tight look, just looking
			/// for further messages from that Buddy. You may then send other 'configuration' messages or,
			/// when you're ready, send another BMSG_HALT message with pMailBox->ArgB2E[MailBoxNum][0]=0,
			/// which will release Emulator from its 'hang' state, resuming normal emulation.

			/// NOTE: Similarly, I/O interface cards can INTERRUPT the Series 80 CPU by setting the LSbit of
			/// the CCR/PSR register. Instead of doing that, Buddies must send a BMSG_INT to Emulator with
			/// pMailBox->ArgB2E[MailBoxNum][0] = the Buddy's I/O "slot number" (see BUDDY.TXT documentation
			/// for the BMSG_INT message, and search for references to 'SlotNum' in this sample Buddy source file).

			/// We take care of IOPS_WANT state right up-front, so Buddy can get his foot in the door (potentially) BEFORE the Emulator is started
			if( IOPstate==IOPS_WANT ) {		/// haven't yet sent WANTS
#if DBGLOG
				OpenLogFile();												/// open BuddyLog file for watching Serial interactions
#endif
				pMailBox->ArgB2E[MailBoxNum][0] = (1<<BWANT_TRACE);			/// set flags for what we WANT
				pMailBox->ArgB2E[MailBoxNum][1] = 060000 | (077777<<16);	/// set START and END addresses for TRACE
				pMailBox->MsgB2E[MailBoxNum] = BMSG_WANT;					/// send the WANT message
				IOPstate = IOPS_HOOK;		/// bump to next state
			}

			if( pMailBox->hWnd[0] ) {	/// if Emulator is running
///===================================================================================
/// before checking for Emulator messages, take care of STATE-specific actions needed.
///===================================================================================
				if( IOPstate >= IOPS_IDLE && (CCR & 0x80) ) IOPstate = IOPS_RESET1;		/// we're being reset

				switch( IOPstate ) {
				 case IOPS_WANT:		/// handled above in order to "get in line" before Emulator (maybe)
					break;
				 case IOPS_HOOK:		/// haven't yet latched into an I/O port
					if( !pMailBox->MsgB2E[MailBoxNum]  ) {		/// if Emulator has taken care of our WANTS message
						pMailBox->MsgB2E[MailBoxNum] = BMSG_IOACTIVE;			/// send msg to find out what I/O addresses are unused
						while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum] ) ;	/// as long as Emulator doesn't terminate, wait until message is processed
						if( pMailBox->hWnd[0] ) {								/// if Emulator is still here
							n = 0106;											/// n = "I/O byte offset" 0177400+0106 = 0177506 = address of CP/M, 2 addresses per S.C. 'slot'
							if( !((pMailBox->BufB2E[MailBoxNum][n>>3]) & (BYTE)(1<<(n&7))) ) {			/// check the bitmask for that I/O slot (1 BIT per I/O address). If unused, we're good to go
								memset((void*)&pMailBox->ArgB2E[MailBoxNum][0], 0, sizeof(pMailBox->ArgB2E[MailBoxNum]));	/// clear all the possible I/O requests
								*(((WORD*)pMailBox->ArgB2E[MailBoxNum])+0) = CCRADR;	/// add the list of I/O addresses we want to 'take'
								*(((WORD*)pMailBox->ArgB2E[MailBoxNum])+1) = IBADR;
								pMailBox->MsgB2E[MailBoxNum] = BMSG_IOTAKE;				/// ask Emulator to hook us into those I/O address
								while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum] ) ;	/// as long as Emulator doesn't terminate, wait until message is processed
								IOPstate = IOPS_IDLE;
							} else {
								MessageBox(NULL, "CP/M I/O addressed already used!", "DEBUG", MB_OK);
								SendMessage(hWndMain, WM_DESTROY, 0, 0);
							}

						} else CloseShop();										/// else reset to "no Emulator" state
					}
					break;
				 case IOPS_RESET1:
					if( !(CCR & 0x80) ) {		/// RESET was released


						/// Do whatever is needed to reset/initialize the CP/M startup process (map ROM into memory, etc.)

						IB = 3;									/// self-test OK
						PSR |= 1;								/// set IBF for CPU80

						IOPstate = IOPS_RESET2;

//						GenInt();							/// generate SVCREQUEST (interrupt CPU80)
					}
					break;
				 case IOPS_RESET2:
					if( !(PSR & 1) ) {						/// if first SELF-TEST-OK byte has been read...
						IB = 3;									/// self-test OK, again
						PSR |= 1;								/// set PED & IBF for CPU80

						IOPstate = IOPS_IDLE;					/// go to IDLE
					}
					break;
				 default:
		///============================================================================================
		/// This is the main pump for dealing with events, sent, and received data (i.e., the IDLE LOOP
		///============================================================================================
					if( IOPstate != IOPS_INT ) {		/// if Emulator is letting us 'run' in IDLE
						HandleOBF();					/// check OBF, deal with it

//						HandleINTofCPU();				/// check if we've interrupted CPU and deal with it (0EB-0FA)
					} else {
						if( !(CCR & 1) ) HandleOBF();					/// check OBF, deal with it (if not being interrupted)
					}
					break;
				}

			}
///===========================================
/// if Emulator has a message for us
///===========================================
			if( pMailBox->MsgE2B[MailBoxNum] && (IOPstate >= IOPS_IDLE || pMailBox->MsgE2B[MailBoxNum]==BMSG_GONE) ) {
	/// NOTE: The "message clearing" (pMailBox->MsgE2B[MailBoxNum] = 0) happens on a case-by-case basis, as some might want to clear it IMMEDIATELY if processing can be 'lengthy' (relatively speaking)
				switch( pMailBox->MsgE2B[MailBoxNum] ) {
				 case BMSG_RESET:
					pMailBox->MsgE2B[MailBoxNum] = 0;									/// release our mailbox for the Emulator to execute this instruction (before sending the next message)
					break;
				 case BMSG_PWON:		/// Emulator is either switching Model OR starting up
					pMailBox->MsgE2B[MailBoxNum] = 0;									/// release our mailbox for the Emulator to execute this instruction (before sending the next message)
					break;
				 case BMSG_PWOFF:		/// Emulator is either switching Model OR terminating
					CloseShop();
					pMailBox->MsgE2B[MailBoxNum] = 0;									/// release our mailbox for the Emulator to execute this instruction (before sending the next message)
					break;
				 case BMSG_GONE:
	/// NOTE: If you don't want your Buddy to quit when the Emulator quits, comment out the following two lines.
					pMailBox->MsgE2B[MailBoxNum] = 0;						/// Clear the message so we don't repeat this during the WM_DESTROY/PostQuitMessage() process
					SendMessage(hWndMain, WM_DESTROY, 0, 0);
					break;
				 case BMSG_TRACE:
#if DBGLOG
					Trace[NextTrace] = LOWORD(pMailBox->ArgE2B[MailBoxNum][0]);	/// save address of this instruction
				 	TraceROM = *(((BYTE*)pMailBox->ArgE2B[MailBoxNum])+7);		/// save ROM# of this instruction
				 	if( ++NextTrace >= TRACELEN ) NextTrace = 0;				/// wrap index around
#endif
					pMailBox->MsgE2B[MailBoxNum] = 0;									/// release our mailbox for the Emulator to execute this instruction (before sending the next message)
					break;
				 case BMSG_IO:
				 	/// NOTE: since we didn't BMSG_WANT with BWANT_IO, we'll ONLY get this message when it's OUR 'hooked' I/O addresses
	/// NOTE: We don't need to deal with INTRSC (177500), as Emulator takes care of that for us
				 	adr = LOWORD(pMailBox->ArgE2B[MailBoxNum][0]);
					if( HIWORD(pMailBox->ArgE2B[MailBoxNum][0])==0x0000FFFF ) {
///vvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
///**** READ
						if( adr==IBADR ) {
	///**** IB
#if DBGLOG
	sprintf(ScratBuf, "R  IB %07o %2.2X", (int)IBADR, (int)(DWORD)IB);
	Log();
#endif
							pMailBox->ArgE2B[MailBoxNum][1] = IB;	/// value to be 'read'
							PSR &= ~(8+4+1);						/// clear PACK, PED, & IBF
						} else {
	///**** PSR
#if DBGLOG
	sprintf(ScratBuf, "R PSR %07o %2.2X", (int)PSRADR, (int)(DWORD)PSR);
	Log();
#endif
							pMailBox->ArgE2B[MailBoxNum][1] = PSR;	/// value to be 'read'
						}
///^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
					} else {
///vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
///**** WRITE
						if( adr==OBADR ) {
	///**** OB
							OB = (BYTE)HIWORD(pMailBox->ArgE2B[MailBoxNum][0]);
							PSR |= 0x80;	/// set OBF
#if DBGLOG
	sprintf(ScratBuf, "W  OB %07o %2.2X", (int)OBADR, (int)(DWORD)OB);
	Log();
#endif
						} else {
	///**** CCR
							b = CCR;						/// save 'previous state' of CCR
							CCR = (BYTE)HIWORD(pMailBox->ArgE2B[MailBoxNum][0]);	/// set 'new state' of CCR
#if DBGLOG
	sprintf(ScratBuf, "W CCR %07o %2.2X", (int)CCRADR, (int)(DWORD)CCR);
	Log();
#endif
/*							/// detect RISING EDGE of INT, set PACK
							if( !(b & 1) && (CCR & 1) ) {				/// if INT was LOW but is going HIGH
								if( (CCR & 1) ) {				/// INT (INTerrupt the IOP), we've caught the RISING EDGE - set PACK in PSR
									PSR |=(0x20 + 0x08);		///			set FDPX & PACK
									if( IOPstate!=IOPS_INT ) {	///
										IOPstate = IOPS_INT;	///		put us in INTERRUPTED state
									} else {					///
										IOPstate = IOPS_IDLE;	///		take us out of INTERRUPTED state
									}
								}
							}
							/// detect FALLING EDGE of INT, clear PACK
							if( (b & 0x01) && !(CCR & 0x01) ) {		/// INT was released
								PSR &= ~8;								/// clear PACK
							}
*/
						}
					}
///^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
					pMailBox->ArgE2B[MailBoxNum][0] = 1;							/// indicate we've handled the access
					pMailBox->MsgE2B[MailBoxNum] = 0;							/// release our mailbox for the Emulator to go again
					break;
				 default:
					pMailBox->MsgE2B[MailBoxNum] = 0;	/// Clear the message to release Emulator for future messages
					break;
				}
			}
		}
///======================================
/// if Windows has a message for us
///======================================
		if( GetMessage(&msg, NULL, 0, 0)==TRUE ) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else break;
	}
    return 0;
}

///*******************************************************************************
LRESULT CALLBACK MainProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {

	HDC				hdc;
	PAINTSTRUCT		ps;
/*	HFONT			oldfont, smallfont;
	RECT			rect;
	HBRUSH			oldbrush;
	HPEN			oldpen;
	int				i, h;
*/
	switch( message ) {
		case WM_PAINT:
			hdc = BeginPaint(hwnd, &ps);
/*
			SetBkColor(hdc, PALETTERGB(255,255,255));
			SetTextColor(hdc, PALETTERGB(0,0,0));

			oldpen = SelectObject(hdc, GetStockObject(BLACK_PEN));
			oldbrush = SelectObject(hdc, GetStockObject(WHITE_BRUSH));

//	Rectangle(hdc, 0, 0, 3000, 3000);

			rect.left = 10;
			rect.top  = 10;
			rect.right= 3000;
			rect.bottom=3000;

			smallfont = CreateFont(-36*48/100, 0, 0,0, FW_BOLD, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, FIXED_PITCH + FF_SWISS, NULL);
			oldfont = (HFONT)SelectObject(hdc, smallfont);
			h = 20;

			for(i=0; i<20; i++) {
				rect.top = 10+i*h;
				if( Keys[i]!=0xFF ) {
					sprintf(ScratBuf, "%s", KeyName[(pMailBox->Model<2)?0:1][Keys[i]]);
					DrawTextEx(hdc, ScratBuf, strlen(ScratBuf), &rect, DT_TOP | DT_LEFT, NULL);
				}
			}
			SelectObject(hdc, oldfont);
			SelectObject(hdc, oldbrush);
			SelectObject(hdc, oldpen);
*/
			EndPaint(hwnd, &ps);
			return 0;
		 case WM_DESTROY:
		 	/// close files and release Emulator hooks
		 	CloseShop();
			/// release the shared memory
        	EndShare(TRUE);
        	/// and quit
            PostQuitMessage (0);       // send a WM_QUIT to the message queue
            return 0;
/*		 case WM_MENUCHAR:
			return (MNC_CLOSE<<16);	// to avoid BEEP when holding ALT down and pressing another key
		 case WM_SYSKEYDOWN:
		 case WM_KEYDOWN:
		 	while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum]!=0 ) ;	/// wait for Emulator to deal with previous message from us
		 	if( pMailBox->hWnd[0] ) {	/// if Emulator is still there
				pMailBox->ArgB2E[MailBoxNum][1] = (BYTE)wParam;		/// set the KEYCOD
				pMailBox->ArgB2E[MailBoxNum][0] = 1;				/// set DOWN
		 pMailBox->MsgB2E[MailBoxNum] = BMSG_KEY;				/// send the KEY message
		 	}
			return 0;
		 case WM_SYSKEYUP:
		 case WM_KEYUP:
		 	while( pMailBox->hWnd[0] && pMailBox->MsgB2E[MailBoxNum]!=0 ) ;	/// wait for Emulator to deal with previous message from us
		 	if( pMailBox->hWnd[0] ) {	/// if Emulator is still there
				pMailBox->ArgB2E[MailBoxNum][1] = (BYTE)wParam;		/// set the KEYCOD
				pMailBox->ArgB2E[MailBoxNum][0] = 0;				/// set UP
				pMailBox->MsgB2E[MailBoxNum] = BMSG_KEY;				/// send the KEY message
		 	}
			return 0;
*/
		 default:
		 	break;
    }

    return DefWindowProc(hwnd, message, wParam, lParam);
}
