/*
 * sdlgraph -- color graphics terminal for Everett Kaser's HP-87 emulator
 *
 * usage: sdlgraph [-d delay] [filename]
 *
 * example:
 *	 sdlgraph -d 1
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <SDL2/SDL.h>
#include <errno.h>

void usage(char *why)
{
	fprintf(stderr,"sdlgraph: Error: %s\n", why);
	printf("usage: sdlgraph [-d delay] [filename]\n");
	exit(-1);
}

#ifdef AZERTY
int french_keyboard(int c, SDL_Keymod mod)
{
	if (mod & KMOD_SHIFT) {
		switch (c) {
		case 249: c = '%'; break;
		case SDLK_EQUALS: c = '+'; break;
		case SDLK_LESS:   c = '>'; break;
		case SDLK_COMMA:  c = '?'; break;
		case SDLK_COLON:  c = '/'; break;
		case SDLK_DOLLAR: c = 30; break;
		case SDLK_ASTERISK: c = 12; break;
		case SDLK_SEMICOLON: c = '.'; break;
		}
	} else {
		if (mod & KMOD_ALT) {
			switch (c) {
			case SDLK_2: c = '~'; break;
			case SDLK_3: c = '#'; break;
			case SDLK_4: c = '{'; break;
			case SDLK_5: c = '['; break;
			case SDLK_6: c = '|'; break;
//			case SDLK_7: c = '`'; break;
			case SDLK_8: c = '\\'; break;
			case SDLK_9: c = '^'; break;
			case SDLK_0: c = '@'; break;
			case SDLK_LEFTPAREN: c = ']'; break;
			case SDLK_EQUALS: c = '}'; break;
			}
		} else switch (c) {
		case 249: c = 'u'; break;
		case SDLK_0: c = 'a'; break;
		case SDLK_1: c = '&'; break;
		case SDLK_2: c = 'e'; break;
		case SDLK_3: c = '"'; break;
		case SDLK_4: c = '\''; break;
		case SDLK_5: c = '('; break;
		case SDLK_6: c = '-'; break;
		case SDLK_7: c = 'e'; break;
		case SDLK_8: c = '_'; break;
		case SDLK_9: c = 'c'; break;
		}
	}
	return c;
}
#endif

/* HP-87 keycodes */
#define HP87_K1		128
#define HP87_K2		129
#define HP87_K3		130
#define HP87_K4		131
#define HP87_K8		132
#define HP87_K9		133
#define HP87_K10	134
#define HP87_K11	135
#define HP87__CHAR	136
#define HP87_CLEAR	137
#define HP87_RUN	141
#define HP87_PAUSE	142
#define HP87_CONT	143
#define HP87_STEP	144
#define HP87_ROLLUP	145
#define HP87_K14	147
#define HP87_LIST	148
#define HP87_PLIST	149
#define HP87_KEYLABEL	150
#define HP87_BACKSPACE	153
#define HP87_ENDLINE	154
#define HP87_FASTBACKSPACE 155
#define HP87_K7		156
#define HP87__LINE	157
#define HP87_I_R	158
#define HP87_LEFT	159
#define HP87_K5		161
#define HP87_K6		162
#define HP87_UP		163
#define HP87_DOWN	164
#define HP87_K12	165
#define HP87_HOME	166	// instead of RESULT
#define HP87_A_G	168
#define HP87_ROLLDOWN	169
#define HP87_RIGHT	170
#define HP87_K13	172

/* actions */
#define PLOT		0
#define CLEAR		1
#define CLEAR_FLAG_NEGATIVE_PEN	0x010000
#define CLEAR_FLAG_GRAPHALL	0x020000
#define CLEAR_FLAG_BYTE_OFFSET	0x040000
#define UNUSED		2
#define COLOR0		3
#define COLOR1		4
#define PIXEL1		5
#define CMD		6
#define CMD_LOWRES		0
#define CMD_HIRES		1
#define CMD_PRESENT		2
#define CMD_CLOSE		3
#define CMD_DELAY		0x26
#define CMD_SYNC		5
#define CMD_ASYNC		6
#define PIXEL0		7
#define DISP		8
#define CRTSAD		9

static char ascii[] = "0123456789";
static char is_on[130560];
static int red   = 0xff;
static int green = 0xff;
static int blue  = 0xff;
static int hires = 0;
static int async = 0;
static int width = 0;
static char *title;
static char *replayfile;
static int x,y;
static int crtsad = 0;
static struct stat statbuf;
static int inputfd;
static int inputlen;
static int outputfd;

#define NONE	 0
#define GRAPH	 1
#define GRAPHALL 2
#define ALPHA	 3
static int window_type = NONE;
static int window_minimized = 0;
static char crt[4320];

#define HEIGHT 240
#define ZOOM 2
static SDL_Renderer *renderer = NULL;
static SDL_Rect point = { 0, 0, ZOOM, ZOOM };

SDL_Window *new_window(int type)
{
	SDL_Window *window;
	int i;

	switch (type) {
	case GRAPH:
		title = "HP-87 GRAPH Screen";
		width = 400;
		for (i = 0; i < width*HEIGHT; i++)
			is_on[i] = 0;
		break;
	case GRAPHALL:
		title = "HP-87 GRAPHALL Screen";
		width = 544;
		for (i = 0; i < width*HEIGHT; i++)
			is_on[i] = 0;
		break;
	case ALPHA:
		title = "HP-87 ALPHA Screen";
		width = 640;
		break;
	}
	window = SDL_CreateWindow("SDL2", SDL_WINDOWPOS_UNDEFINED,
				  SDL_WINDOWPOS_UNDEFINED, width*ZOOM,
				  HEIGHT*ZOOM, SDL_WINDOW_SHOWN);
	SDL_SetWindowTitle(window, title);
	renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_SOFTWARE);
	return window;
}

SDL_Window *default_window()
{
	SDL_Window *window;

	/*
	 * create a default color GRAPH NORMAL window
	 */
	window_type = GRAPH;
	window = new_window(window_type);
	SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0xff);
	SDL_RenderClear(renderer);
	SDL_SetRenderDrawColor(renderer, red, green, blue, 0xff);
	SDL_RenderPresent(renderer);
	return window;
}

SDL_Window *clear_window(SDL_Window *window, int n)
{
	int start_line;
	int end_line;
	int on;
	int i;
	int clear_request;

	if (n & CLEAR_FLAG_GRAPHALL)
		clear_request = GRAPHALL;
	else
		clear_request = GRAPH;
	if (window_type != clear_request) {
		/*
		 * switch between GRAPH and GRAPHALL
		 * let's substitute a new window to the old one
		 */
		if (window) {
			SDL_DestroyRenderer(renderer);
			SDL_DestroyWindow(window);
		}
		window_type = clear_request;
		window = 0;
	}
	if (!window)
		window = new_window(window_type);
	if (n & CLEAR_FLAG_BYTE_OFFSET) {
		start_line = ((n & 0x00ffff) << 3) / width;
		end_line = 239;
	} else {
		start_line = n & 0x0000ff;
		end_line = (n & 0x00ff00) >> 8;
	}
	on = (n & CLEAR_FLAG_NEGATIVE_PEN) >> 16;
	on = !!on;
	for (i = start_line*width; i < (end_line + 1)*width; i++)
		is_on[i] = on;
	if (on)
		SDL_SetRenderDrawColor(renderer, red, green, blue, 0xff);
	else
		SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0xff);
	point.x = 0;
	point.y = start_line*ZOOM;
	point.w = width*ZOOM;
	point.h = (end_line - start_line + 1)*ZOOM;
	SDL_RenderFillRect(renderer, &point);
	point.w = ZOOM;
	point.h = ZOOM;
	if (!async)
		SDL_RenderPresent(renderer);
	if (!on)
		SDL_SetRenderDrawColor(renderer, red, green, blue, 0xff);
	return window;
}

void process_bit(int bit, int on)
{
	on = !!on;
	if (hires && (is_on[bit] == on))
		return;
	is_on[bit] = on;
	point.x = (bit % width)*ZOOM;
	point.y = (bit / width)*ZOOM;
	SDL_SetRenderDrawColor(renderer, on*red, on*green, on*blue, 0xff);
	SDL_RenderFillRect(renderer, &point);
}

void handle_sdl_events()
{
	SDL_Event event;
	int c, rc;

	if (!SDL_PollEvent(&event))
		return;
	if (event.type == SDL_QUIT)
		exit(0);
	if (event.type == SDL_WINDOWEVENT) {
		if (event.window.event == SDL_WINDOWEVENT_CLOSE)
			exit(0);
		else if (event.window.event == SDL_WINDOWEVENT_MINIMIZED)
			window_minimized = 1;
		else if (event.window.event == SDL_WINDOWEVENT_RESTORED)
			window_minimized = 0;
		if (renderer)
			SDL_RenderPresent(renderer);
	} else if (!replayfile && event.type == SDL_KEYDOWN) {
		c = event.key.keysym.sym;
		// fprintf(stderr,"==%d==\n",c);
		switch (c) {
		case SDLK_ESCAPE:	c = HP87_KEYLABEL; break;
		case SDLK_RETURN:	c = HP87_ENDLINE; break;
		case SDLK_RIGHT:	c = HP87_RIGHT; break;
		case SDLK_LEFT:		c = HP87_LEFT; break;
		case SDLK_DOWN:		c = HP87_DOWN; break;
		case SDLK_UP:		c = HP87_UP; break;
		case SDLK_PAGEDOWN:	c = HP87_ROLLDOWN; break;
		case SDLK_PAGEUP:	c = HP87_ROLLUP; break;
		case SDLK_INSERT:	c = HP87_I_R; break;
		case SDLK_F10:		c = HP87_A_G; break;
		}
		switch (c) {
		case SDLK_LALT:		break;	/* ignore */
		case SDLK_RALT:		break;	/* ignore */
		case SDLK_LCTRL:	break;	/* ignore */
		case SDLK_RCTRL:	break;	/* ignore */
		case SDLK_LSHIFT:	break;	/* ignore */
		case SDLK_RSHIFT:	break;	/* ignore */
		case SDLK_TAB:		break;	/* ignore */
		default:
			if (event.key.keysym.mod & KMOD_SHIFT) {
				if (SDLK_a <= c && c <= SDLK_z)
					c = c - 32;
				else switch (c) {
				case SDLK_F1:	c = HP87_K8; break;
				case SDLK_F2:	c = HP87_K9; break;
				case SDLK_F3:	c = HP87_K10; break;
				case SDLK_F4:	c = HP87_K11; break;
				case SDLK_F5:	c = HP87_K12; break;
				case SDLK_F6:	c = HP87_K13; break;
				case SDLK_F7:	c = HP87_K14; break;
				case SDLK_F8:	c = HP87_CONT; break;
				case SDLK_F9:	c = HP87_PLIST; break;
				case SDLK_HOME:	c = HP87_CLEAR; break;
				case SDLK_END:
				case SDLK_PAUSE: c = HP87_STEP; break;
				case SDLK_DELETE: c = HP87__LINE; break;
				case SDLK_BACKSPACE: c = HP87_FASTBACKSPACE; break;
				}
			} else {
				switch (c) {
				case SDLK_F1:	c = HP87_K1; break;
				case SDLK_F2:	c = HP87_K2; break;
				case SDLK_F3:	c = HP87_K3; break;
				case SDLK_F4:	c = HP87_K4; break;
				case SDLK_F5:	c = HP87_K5; break;
				case SDLK_F6:	c = HP87_K6; break;
				case SDLK_F7:	c = HP87_K7; break;
				case SDLK_F8:	c = HP87_RUN; break;
				case SDLK_F9:	c = HP87_LIST; break;
				case SDLK_HOME: c = HP87_HOME; break;
				case SDLK_END:
				case SDLK_PAUSE: c = HP87_PAUSE; break;
				case SDLK_DELETE: c = HP87__CHAR; break;
				case SDLK_BACKSPACE: c = HP87_BACKSPACE; break;
				}
			}
#ifdef AZERTY
			c = french_keyboard(c, event.key.keysym.mod);
#endif
			switch (c) {
				case SDLK_KP_0: c = '0'; break;
				case SDLK_KP_1: c = '1'; break;
				case SDLK_KP_2: c = '2'; break;
				case SDLK_KP_3: c = '3'; break;
				case SDLK_KP_4: c = '4'; break;
				case SDLK_KP_5: c = '5'; break;
				case SDLK_KP_6: c = '6'; break;
				case SDLK_KP_7: c = '7'; break;
				case SDLK_KP_8: c = '8'; break;
				case SDLK_KP_9: c = '9'; break;
				case SDLK_KP_PERIOD: c = '.'; break;
				case SDLK_KP_DIVIDE: c = '/'; break;
				case SDLK_KP_MULTIPLY: c = '*'; break;
				case SDLK_KP_PLUS: c = '+'; break;
				case SDLK_KP_MINUS: c = '-'; break;
				case SDLK_KP_ENTER: c = HP87_ENDLINE; break;
//				case SDLK_NUMLOCKCLEAR: c = ????; break;
			}
			if (event.key.keysym.mod & KMOD_CTRL) {
				if (!(event.key.keysym.mod & KMOD_SHIFT)) {
					if (SDLK_a <= c && c <= SDLK_z)
						c = c - 32;
				}
				if (SDLK_AT <= c && c <= SDLK_UNDERSCORE)
					c = c - 64;
			}
			rc = write(outputfd, &c, 1);
			break;
		}
	}
}

void check_rewind()
{
	fstat(inputfd, &statbuf);
	if (inputlen > statbuf.st_size) {
		inputlen = 0;			/* file has been truncated */
		lseek(inputfd, 0, SEEK_SET);	/* let's rewind */
	} else
		inputlen = statbuf.st_size;
}

int follow_read(char *cp)
{
	int rc;

	handle_sdl_events();

	/* do not modify window while it is minimized */
	while (window_minimized) {
		SDL_Delay(100);
		handle_sdl_events();
	}
	rc = read(inputfd, cp, 1);
	if (rc < 0) {
		perror("sdlgraph: Error reading input file");
		exit(rc);
	}
	if (!rc)
		check_rewind();
	return rc;
}

int main(int argc, char *argv[])
{
	int on, rc;
	int addr, bit, i;
	char c1, c2, c3, c4, c5;
	char in[17] = "0123456789abcdef";
	SDL_Window *window = NULL;
	char *title, *s;
	unsigned int n;
	int delay = 0;
	int count = 0;

	i = 1;
	if (argc > 1) {
#ifdef __linux__
		if (argv[1][0]=='-')
#else
		if (argv[1][0]=='-' || argv[1][0]=='/')
#endif
		{
			if (argv[1][1] != 'd')
				usage("Unknow flag");
			if (argc == 2)
				usage("Missing delay value");
			if (argv[2][0] < '0' || argv[2][0] > '9')
				usage("Non-numeric delay value");
			delay = atoi(argv[2]);
			i = 3;
		}
	}
	if (argc > i+1)
		usage("Too many arguments");
	replayfile = argv[i];
	s = replayfile ? replayfile : "PRT710.TXT";
	inputfd = open(s, O_RDONLY);
	if (inputfd == -1) {
		fprintf(stderr, "sdlgraph: Error opening input file %s: ", s);
		perror("");
		exit(-1);
	}
	fstat(inputfd, &statbuf);
	inputlen = statbuf.st_size;
	/* discard keyboard input during replay */
	if (!replayfile) {
		outputfd = open("SDKEYS.TXT", O_WRONLY|O_CREAT|O_APPEND,
				  S_IRUSR|S_IWUSR);
		if (outputfd == -1) {
			perror("sdlgraph: Error opening output file SDKEYS.TXT");
			exit(-1);
		}
	}
	rc = SDL_Init(SDL_INIT_VIDEO);
	if (rc == -1) {
		fprintf(stderr, "sdlgraph: SDL_Init(): %s\n", SDL_GetError());
		exit(-1);
	}
	for (i = 0; i < 4320; i++)
		crt[i] = ' ';
	window = default_window();
	/*
	 * Read messages from input file, and decode them.
	 * Messages are 5 byte long, followed by \n or \r\n.
	 *
	 * Bytes c1 to c4 hold a series80-style BCD-encoded number:
	 *   - The first seven four-bit digits represent the value,
	 *     little endian, left-justified (0 to 9999999).
	 *   - The last four-bit digit specifies the number of
	 *     significant digits minus 1 (0 to 6).
	 * Byte c5 defines the associated action to be performed.
	 */
	while (1) {
		if (delay) {
#ifdef __linux__
			usleep(delay); /* microseconds */
#else
			count = ++count % 100;
			if (!count)
				SDL_Delay(delay); /* milliseconds */
#endif
		}
		do { rc = follow_read(&c1); } while (rc != 1);
		if (c1 == '\n') continue;
		do { rc = follow_read(&c2); } while (rc != 1);
		if (c2 == '\n') continue;
		do { rc = follow_read(&c3); } while (rc != 1);
		if (c3 == '\n') continue;
		do { rc = follow_read(&c4); } while (rc != 1);
		if (c4 == '\n') continue;
		do { rc = follow_read(&c5); } while (rc != 1);
		if (c5 == '\n') continue;
		if (c5 == '\r') continue;

		in[0] = ascii[(c1 & 0xf0) >> 4];
		in[1] = ascii[c1 & 0x0f];
		in[2] = ascii[(c2 & 0xf0) >> 4];
		in[3] = ascii[c2 & 0x0f];
		in[4] = ascii[(c3 & 0xf0) >> 4];
		in[5] = ascii[c3 & 0x0f];
		in[6] = ascii[(c4 & 0xf0) >> 4];
		in[(c4 & 0x0f) + 1] = '\n';

		rc = sscanf(in, "%d", &n);

		switch (c5) {
		case CRTSAD:
			addr = (n & 0xffff00) >> 8;	/* address */
			n &= 0xff;
			fprintf(stderr,"CRTSAD addr=0x%x ", addr);
			fprintf(stderr,"%s\n", n ? "CLEAR" : "SET");
			if (addr != crtsad) {
				fprintf(stderr, "scroll\n");
				crtsad = addr;
			}
//			if (!async)
//				SDL_RenderPresent(renderer);
			break;
		case DISP:
			addr = (n & 0xffff00) >> 8;	/* address */
			n &= 0xff;
			crt[addr] = n;
			fprintf(stderr,"DISP addr=0x%x curs=%d ", addr, !!(n & 0x80));
			if ((n & 0x7f) == 0x0d)
				fprintf(stderr,"'\\n'\n");
			else
				fprintf(stderr,"'%c'\n", n & 0x7f);
			fprintf(stderr,"DISP y=%d x=%d\n", addr/80*15, addr%80);
//			if (!async)
//				SDL_RenderPresent(renderer);
			break;
		case PLOT:
			bit = (n & 0xffff00) >> 5;	/* bit offset */
			process_bit(bit++, n & 0x000080);
			process_bit(bit++, n & 0x000040);
			process_bit(bit++, n & 0x000020);
			process_bit(bit++, n & 0x000010);
			process_bit(bit++, n & 0x000008);
			process_bit(bit++, n & 0x000004);
			process_bit(bit++, n & 0x000002);
			process_bit(bit++, n & 0x000001);
			if (!async)
				SDL_RenderPresent(renderer);
			SDL_SetRenderDrawColor(renderer, red, green, blue, 0xff);
			break;
		case CLEAR:
			window = clear_window(window, n);
			break;
		case COLOR1:
			n |= 0x800000;		/* set most significant bit */
			/* FALL THROUGH */
		case COLOR0:
			red = (n & 0xff0000) >> 16;
			green = (n & 0x00ff00) >> 8;;
			blue = n & 0x0000ff;
			if (window)
				SDL_SetRenderDrawColor(renderer,
						       red, green, blue, 0xff);
			break;
		case PIXEL0:
		case PIXEL1:
			y = n & 0x0000ff;		/* bit y coordinate */
			x = (n & 0xffff00) >> 8;	/* bit x coordinate */
			bit = x + y*width;		/* bit offset */
			point.x = x*ZOOM;
			point.y = y*ZOOM;
			on = is_on[bit];
			if (c5 == PIXEL1) {
				is_on[bit] = 1;
				SDL_RenderFillRect(renderer, &point);
			} else {
				is_on[bit] = 0;
				SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0xff);
				SDL_RenderFillRect(renderer, &point);
				SDL_SetRenderDrawColor(renderer,
						       red, green, blue, 0xff);
			}
			if (!async)
				SDL_RenderPresent(renderer);
			break;
		case CMD:
			switch (c1) {
			case CMD_LOWRES:
				hires = 0;
				break;
			case CMD_HIRES:
				hires = 1;
				break;
			case CMD_PRESENT:
				if (!async)
					break;
				if (window)
					SDL_RenderPresent(renderer);
				break;
			case CMD_CLOSE:
				exit(0);
				break;
			case CMD_DELAY:
				n = (n & 0xff)*10;	/* 10th of seconds */
				while (n) {
					SDL_Delay(100);	/* ms */
					handle_sdl_events();
					/* stop countdown while window is minimized */
					if (!window_minimized)
						n--;
				}
				break;
			case CMD_SYNC:
				async = 0;
				break;
			case CMD_ASYNC:
				async = 1;
				break;
			default:
				//fprintf(stderr, "c1=%d c5=%d\n", c1, c5);
				break;
			}
			break;
		default:
			//fprintf(stderr, "c1=%d c2=%d c3=%d c4=%d c5=%d\n",
			//	c1, c2, c3, c4, c5);
			break;
		}
	}
}

#ifdef NOTYET
static char font87[128][8] = {
	{4,12,28,60,28,12,4,0},
	{8,0,8,16,34,34,28,0},
	{62,0,34,20,8,20,34,0},
	{62,0,34,50,42,38,34,0},
	{0,0,26,36,36,36,26,0},
	{28,34,34,60,34,34,60,0},
	{62,34,16,8,16,34,62,0},
	{0,8,8,20,20,34,62,0},
	{0,8,16,62,16,8,0,0},
	{0,0,30,36,36,36,24,0},
	{8,28,42,8,8,8,8,0},
	{32,16,8,20,34,34,34,0},
	{0,0,18,18,18,28,16,32},
	{0,0,0,0,0,0,0,0},
	{0,2,28,40,8,8,8,0},
	{0,2,28,52,20,20,20,0},
	{28,34,34,62,34,34,28,0},
	{30,40,40,60,40,40,46,0},
	{0,0,20,42,46,40,22,0},
	{8,28,34,34,62,34,34,0},
	{8,0,28,36,36,36,26,0},
	{20,28,34,34,62,34,34,0},
	{20,0,28,36,36,36,26,0},
	{20,28,34,34,34,34,28,0},
	{20,0,28,34,34,34,28,0},
	{20,34,34,34,34,34,28,0},
	{20,0,34,34,34,34,28,0},
	{32,32,32,62,32,32,32,0},
	{8,8,8,8,8,8,8,8},
	{0,0,0,255,0,0,0,0},
	{12,18,16,56,16,16,62,0},
	{8,8,8,255,8,8,8,8},
	{0,0,0,0,0,0,0,0},
	{8,8,8,8,8,0,8,0},
	{20,20,20,0,0,0,0,0},
	{20,20,62,20,62,20,20,0},
	{8,30,40,28,10,60,8,0},
	{48,50,4,8,16,38,6,0},
	{16,40,40,16,42,36,26,0},
	{8,8,8,0,0,0,0,0},
	{4,8,16,16,16,8,4,0},
	{16,8,4,4,4,8,16,0},
	{8,42,28,8,28,42,8,0},
	{0,8,8,62,8,8,0,0},
	{0,0,0,0,0,8,8,16},
	{0,0,0,62,0,0,0,0},
	{0,0,0,0,0,8,8,0},
	{0,2,4,8,16,32,0,0},
	{28,34,38,42,50,34,28,0},
	{8,24,8,8,8,8,28,0},
	{28,34,2,12,16,32,62,0},
	{62,2,4,12,2,34,28,0},
	{4,12,20,36,62,4,4,0},
	{62,32,60,2,2,34,28,0},
	{14,16,32,60,34,34,28,0},
	{62,2,4,8,16,16,16,0},
	{28,34,34,28,34,34,28,0},
	{28,34,34,30,2,4,56,0},
	{0,0,8,8,0,8,8,0},
	{0,0,8,8,0,8,8,16},
	{4,8,16,32,16,8,4,0},
	{0,0,62,0,62,0,0,0},
	{32,16,8,4,8,16,32,0},
	{28,34,34,4,8,0,8,0},
	{28,34,42,46,44,32,30,0},
	{28,34,34,62,34,34,34,0},
	{60,34,34,60,34,34,60,0},
	{28,34,32,32,32,34,28,0},
	{60,34,34,34,34,34,60,0},
	{62,32,32,60,32,32,62,0},
	{62,32,32,60,32,32,32,0},
	{28,34,32,32,38,34,30,0},
	{34,34,34,62,34,34,34,0},
	{28,8,8,8,8,8,28,0},
	{2,2,2,2,2,34,28,0},
	{34,36,40,48,40,36,34,0},
	{32,32,32,32,32,32,62,0},
	{34,54,42,34,34,34,34,0},
	{34,50,50,42,38,38,34,0},
	{28,34,34,34,34,34,28,0},
	{60,34,34,60,32,32,32,0},
	{28,34,34,34,42,36,26,0},
	{60,34,34,60,40,36,34,0},
	{28,34,32,28,2,34,28,0},
	{62,8,8,8,8,8,8,0},
	{34,34,34,34,34,34,28,0},
	{34,34,34,34,20,20,8,0},
	{34,34,34,34,42,54,34,0},
	{34,20,20,8,20,20,34,0},
	{34,20,20,8,8,8,8,0},
	{62,4,4,8,16,16,62,0},
	{62,32,32,32,32,32,62,0},
	{0,32,16,8,4,2,0,0},
	{62,2,2,2,2,2,62,0},
	{8,20,34,0,0,0,0,0},
	{0,0,0,0,0,0,62,0},
	{32,16,8,0,0,0,0,0},
	{0,0,28,2,30,34,30,0},
	{32,32,60,34,34,34,60,0},
	{0,0,28,34,32,34,28,0},
	{2,2,30,34,34,34,30,0},
	{0,0,28,34,62,32,30,0},
	{12,18,16,56,16,16,16,0},
	{0,0,30,34,34,30,2,28},
	{32,32,60,34,34,34,34,0},
	{8,0,24,8,8,8,28,0},
	{4,0,12,4,4,4,20,8},
	{16,16,18,20,24,20,18,0},
	{24,8,8,8,8,8,28,0},
	{0,0,52,42,42,42,42,0},
	{0,0,60,34,34,34,34,0},
	{0,0,28,34,34,34,28,0},
	{0,0,60,34,34,60,32,32},
	{0,0,28,36,36,28,4,6},
	{0,0,60,34,32,32,32,0},
	{0,0,30,32,28,2,60,0},
	{0,8,28,8,8,8,4,0},
	{0,0,34,34,34,34,30,0},
	{0,0,34,18,20,12,8,0},
	{0,0,34,34,42,42,20,0},
	{0,0,34,20,8,20,34,0},
	{0,0,34,34,34,30,2,12},
	{0,0,62,4,8,16,62,0},
	{4,8,8,16,8,8,4,0},
	{8,8,8,8,8,8,8,0},
	{16,8,8,4,8,8,16,0},
	{0,0,16,42,4,0,0,0},
	{42,20,42,20,42,20,42,0},
};
#endif
